/*
 * Functions for elementary operations for
 * simplicial complex computations.
 *
 * by M.Hachimori, May 2000.
 */

#include <stdlib.h>
#include "s_complex.h"



/*
 * Function to enumerate all faces from
 * a list of facets.
 * The input list of facets are assumed to be sorted.
 * The output list is also sorted.
 */
Simplices *facets2faces(faces)
  Simplices *faces;
{
  Simplices *initial, *tail, *current, *output, *return_address, *tmp, *tmp2;
  int i, j, d, num_newface;

  if (faces == NULL)
    return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  return_address = output;
  output->prev = NULL;
  output->next = NULL;
  initial = output; /* 'initial' be the first and 'tail' be the last data of new d-dim faces.*/
  d = faces->vert[0];

  current = faces;    /* 'current' is a pointer sweeping the input data. */

  /* copy the largest dimension part */
  while(current != NULL && current->vert[0] == d){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output = output->next;
    for (i = 0; i <= current->vert[0]+1; i++)
      output->vert[i] = current->vert[i];
    output->next = NULL;
    current = current->next;
    }
  tail = output;
  d--;
  initial = initial->next; 

  while(output->vert[0] >= 1){
    num_newface = 0;
    /* copy 1-dim-lower faces to output.*/
    while (current!= NULL && current->vert[0] > d)
      current = current->next;
    while (current!= NULL && current->vert[0] == d){
      output->next = (Simplices*)malloc(sizeof(Simplices));
      for (i = 0; i <= current->vert[0]+1; i++)
        output->next->vert[i] = current->vert[i];
      output->next->prev = output;
      output = output->next;
      output->next = NULL;
      num_newface ++;
      current = current->next;}

  
    /* compute codim 1 faces of faces between `initial' and `tail'
       and add them to output. */
    while(initial != NULL && initial->vert[0] == d+1){
      for (i = 1; i <= initial->vert[0]+1; i ++){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output = output->next;
        output->next = NULL;
        output->vert[0] = initial->vert[0] - 1;
        for (j = 1; j <= initial->vert[0]+1; j++) /* list all codim 1 faces */
          if (j < i) output->vert[j] = initial->vert[j];
          else if (j > i) output->vert[j-1] = initial->vert[j];
        num_newface ++;}
      initial = initial->next;
      }

    /* sort the faces and remove multiplicity before going to the next step*/
    initial = tail->next;
    msort(initial, num_newface);
    tmp = initial;
    while(tmp->next != NULL){
      if (cmp_faces(tmp->vert, (tmp->next)->vert) == 0){
        tmp2 = (tmp->next)->next;
        free(tmp->next);
        tmp->next = tmp2;
        if(tmp->next != NULL)
          tmp->next->prev = tmp;
        }
      else
        tmp = tmp->next;}
    tail = tmp;
    output = tmp;

    d --;
    }

  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;
  return return_address;
}


/*
 * Function to calculate a set of facets
 * from a set of faces.
 */
Simplices *faces2facets(faces)
  Simplices *faces;
{
  Simplices *return_address, *head, *tail, *current, *output, *tmp;
  int i, j, d, mark;

  if (faces == NULL) return NULL;


  d = faces->vert[0];
  output = (Simplices*)malloc(sizeof(Simplices));
  output->prev = NULL;
  output->next = NULL;
  return_address = output;

  current = faces;
  head = current;
  /* All the largest dimensional faces are copied to output. */
  while(current!= NULL && current->vert[0] == d){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output = output->next;
    output->next = NULL;
    for (i = 0; i <= current->vert[0]+1; i ++)
      output->vert[i] = current->vert[i];
    current = current->next;
    }
  d --;

  /* For the smaller dimensional faces, check whether they are
     contained in a larger face or not. */
  while(d >= 0){
    while(current != NULL && current->vert[0] == d) {
      tmp = head;
      mark = 0;
      while(tmp != NULL && tmp->vert[0] > d) {
        if (face_contained(current->vert, tmp->vert) == 1){
          mark = 1;
          break;}
        tmp = tmp->next;
        }
      if (mark == 0){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output = output->next;
        output->next = NULL;
        for (i = 0; i <= current->vert[0]+1; i ++)
          output->vert[i] = current->vert[i];
        }
      current = current->next;
      }

    d --;
  }

  /* remove the first extra data before returning. */
  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;
  return return_address;
}



/*
 * Function to calculate boundary complex
 *   the input is a list of facets and
 *   the output is also a list of facets.
 */
Simplices *boundary_facets(facets)
  Simplices *facets;
{
  Simplices *return_address, *output, *head;
  Simplices *tmp, *tmp1, *tmp2;
  Simplices *top_added, *current1, *current2;
  int i, j, d, num_newface;

  if (facets == NULL || facets->vert[0] <= 0)
     return NULL;

  output = (Simplices*)malloc(sizeof(Simplices));
  output->next = NULL;
  return_address = output;
          /* This first data will not be used.
             Just defined here to adjust for the looping
             and removed in the last of this function. */
  d = facets->vert[0];

  while(d > 0){
    head = output;
    num_newface = 0;
    /* first list up all codim 1 faces of d-dim facets. */
    while(facets != NULL && facets->vert[0] == d){
      for (i = 1; i <= facets->vert[0]+1; i ++){
        output->next = (Simplices*)malloc(sizeof(Simplices));
        output->next->prev = output;
        output = output->next;
        output->next = NULL;
        output->vert[0] = facets->vert[0]-1;
        for (j = 1; j <= facets->vert[0]+1; j ++)
          if (j < i) output->vert[j] = facets->vert[j];
          else if (j > i) output->vert[j-1] = facets->vert[j];
        num_newface ++;
        }
      facets = facets->next;}
    msort(head->next, num_newface);

    /* if there are more than one identical faces, then
       remove all of them.*/
    top_added = head->next;
    while(head->next != NULL){
      if (cmp_faces(head->vert, head->next->vert) == 0)
        if (head->next->next != NULL &&
            cmp_faces(head->next->vert, head->next->next->vert) == 0){
          tmp = head->next->next;
          free(head->next);
          head->next = tmp;
          head->next->prev = head;}
        else
          if (head->prev != NULL){
            tmp1 = head->prev;
            tmp2 = head->next->next;
            free(head->next);
            free(head);
            if (tmp1 != NULL) tmp1->next = tmp2;
            if (tmp2 != NULL) tmp2->prev = tmp1;
            head = tmp2;
            if (head == NULL) break;}
          else{
            tmp2 = head->next->next;
            free(head->next);
            free(head);
            tmp2->prev = NULL;
            head = tmp2;}
      else
        head = head->next;
      }

    /* if d is not the largest dimension, then remove
       faces which is contained in other larger face. */
    current1 = top_added; /* top_added is the top face of added faces.*/
    while(current1 != NULL){
      current2 = return_address->next; /* current2 sweeps all the larger
                                         faces in the output. */
      if (current2 == NULL) break; /* if no boundary, break. */
      while(current2->vert[0] >= d){
        if (face_contained(current1->vert, current2->vert) == 1){
          tmp = current1->prev;
          tmp->next = current1->next;
          current1->next->prev = tmp;
          free(current1);
          current1 = tmp->next;
          continue;}
        current2 = current2->next;}
      current1 = current1->next;
      }

    d --;
  }

  /* remove the null head */
  if (return_address->next != NULL){
    return_address = return_address->next;
    free(return_address->prev);
    return_address->prev = NULL;
    return return_address;}
  else{
    free(return_address);
    return NULL;}

}


/*
 * Function to calculate k-th pure part.
 *   Input = set of faces
 *   Output = set of facets of k-th pure part
 */
Simplices *pure_part(faces, k)
  Simplices *faces;
  int k;
{
  Simplices *output, *return_address;
  int i;

  if (faces == NULL || faces->vert[0] < k)
    return NULL;

  output = (Simplices*)malloc(sizeof(Simplices)); /* this first data will not
                                                     be used and will be removed
                                                     in the last of this function.*/
  output->prev = NULL;
  return_address = output;

  /* search the first face whose dimension is k */
  while(faces->vert[0] > k)
    faces = faces->next;

  /* output all k-dim faces */
  while(faces != NULL && faces->vert[0] == k){
    output->next = (Simplices*)malloc(sizeof(Simplices));
    output->next->prev = output;
    output->next->next = NULL;
    output = output->next;
    for (i = 0; i <= faces->vert[0]+1; i ++)
      output->vert[i] = faces->vert[i];
    faces = faces->next;
    }

  /* remove the first extra data */
  return_address = return_address->next;
  free(return_address->prev);
  return_address->prev = NULL;

  return return_address;

}


/*
 * Function to calculate f-vector.
 *   Input is a set of faces
 *   Output is a vector whose i-th element is f_(i-1)
 *   (Note that the index is shifted by 1.
 *    If the input is NULL, then return f_vec[0] = 1,
 *    just show that there is only one (-1)-dim face.)
 */
void get_fvector(faces, f_vec)
  Simplices *faces;
  int f_vec[];
{
  int i;

  if (faces == NULL){
    f_vec[0] = 1;
    return;}

  f_vec[0] = 1;
  /* initially set f_vec[i] = 0 for all i>=1. */
  for (i = 1; i <= faces->vert[0]+1; i ++)
    f_vec[i] = 0;

  /* count the number of faces. */
  while(faces != NULL){
    f_vec[faces->vert[0]+1] ++;
    faces = faces->next;
    }

  return;
}


/*
 * Function to calculate h-vector from f-vector.
 */
void f_to_h(fvec, hvec, dim)
  int fvec[];
  int hvec[];
  int dim;
{
  int i, j;
  int tmp[MAXDIM+2], tmp2[MAXDIM+2];

  tmp[0] = fvec[0];
  for (i = 0; i <= dim+1; i ++)
    {
    tmp2[0] = 1;
    for (j = 1; j <= i; j ++)
      tmp2[j] = tmp[j] - tmp[j - 1];
    tmp2[i + 1] = fvec[i + 1];
    for (j = 0; j <= i + 1; j ++)
      tmp[j] = tmp2[j];
    }
  for(i = 0; i <= dim+1; i ++)
    hvec[i] = tmp[i];
  return;
}


/*
 * Function to check whether a vector is an M-sequence or not.
 *   dim = dimension = (number of entries -1)
 */
int Mseq(hvec, dim)
  int hvec[];
  int dim;
{
  short i;


  if (hvec[0] != 1)
    return 0;
  for (i = 1; i <= dim; i ++)
    if (hvec[i] < 0 || (hvec[i] > 0 && M(hvec[i], i) < hvec[i + 1]))
      return 0;
  if (hvec[dim+1] < 0)
    return 0;
  return 1;
}

/*
 * function used in Mseq.
 */
int M(n, i)
  int n;
  int i;
{
  int j, k;
  int ans, bi[MAXDIM+1];

  binom_expand(n, i, bi);
  k = 1;
  while(bi[k] == 0)
    k ++;
  ans = 0;
  for (j = k; j <= i; j ++)
    ans += binom(bi[j] + 1, j + 1);

  return ans;
}


/*
 * function to calculate binomial coefficient
 */
int binom(n, r)
  int n;
  int r;
{
  int i;
  double ans;

  ans = 1;
  for (i = n - r + 1; i <= n; i ++)
    ans *= i;
  for (i = 2; i <= r; i ++)
    ans /= i;

  return (int)ans;
}


/*
 * function to calculate binomial expansion
 */
void binom_expand(n, i, bi)
  int n;
  int i;
  int bi[];
{
  int j;
  int tmp;

  while(n > 0)
    {
    tmp = i;
    while (binom(tmp + 1, i) <= n)
      tmp ++;
    bi[i] = tmp;
    n -= binom(tmp, i);
    i --;
    }
  for (j = 1; j <= i; j ++)
    bi[j] = j-1;
  bi[0] = 0;

  return;
}
